/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#ifdef __FreeBSD__
#include <unistd.h>
#include <netinet/in.h>
#include <sys/socket.h>
#endif

#include <arpa/inet.h>
#include <netdb.h>
#include <pwd.h>
#include "pan.h"

#include "../config.h"

#ifdef AF_NEEDS_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#include "grouplist.h"
#include "nntp.h"
#include "queue.h"
#include "queue-item-send.h"


static const gchar* get_host_name (void);
static const gchar* get_fqdn (const gchar *host);
static gchar* generate_references (Message *msg);

#if 0
static gint
check_from_syntax (gchar *s)
{
	return 1;
}
#endif


#if 0
static gint
check_newsgroups_syntax (gchar *s)
{
	return 1;
}
#endif

typedef enum
{
	OKAY,
	WARN,
	REFUSE,
}
GoodnessLevel;

static gchar*
check_body (const gchar* body, GoodnessLevel *goodness)
{
	gchar *pch = NULL;
	gchar *tmp = NULL;
	gchar **sd = NULL;
	int i = 0;
	int lines = 0;
	int unquoted_lines = 0;

	g_assert (goodness!=NULL);

	/* check for an empty message */
	if (!body || !*body) {
		tmp = g_strjoin ("\n", _("The message is empty."), pch, NULL);
		g_free (pch);
		pch = tmp;
		if (*goodness<REFUSE)
			*goodness=REFUSE;
	}

	/* check number of new lines compared to quoted lines */
	unquoted_lines = lines = 0;
	sd = g_strsplit (body, "\n", -1);
	for (i=0; sd[i]!=NULL; ++i) {
		++lines;
		if (*sd[i]!='>')
			++unquoted_lines;
	}
	g_strfreev (sd);
	if (!unquoted_lines) {
		tmp = g_strjoin ("\n",
			_("The message is entirely quoted text."), pch, NULL);
		g_free (pch);
		pch = tmp;
		if (*goodness<REFUSE)
			*goodness=REFUSE;
	}
	else if ((unquoted_lines/(double)lines)<0.25)
	{
		tmp = g_strjoin ("\n",
		       	_("The message is mostly quoted text."), pch, NULL);
		g_free (pch);
		pch = tmp;
		if (*goodness<WARN)
			*goodness=WARN;
	}

	return pch;
}


static gchar *
generate_references (Message *msg)
{
#define GNKSA_LENGTH 998
	gchar buf[GNKSA_LENGTH + 1];
	gchar *p = NULL;
	gchar *refs = NULL;
	gchar *new_refs = NULL;
	gchar *pch = NULL;
	unsigned int total_len, len;

//	p = extract_header (mw->article, "References");
//	if (p)
//		refs = g_strconcat (p, " ", extract_header(mw->article, "Message-ID"), NULL);
//	else
//		refs = g_strconcat (extract_header(mw->article, "Message-ID"), NULL);
//	g_free (p);
	
	if (msg->references) {
		refs = g_strconcat (msg->references, " ", Pan.active_message_id, NULL);
	}
	else if (Pan.active_message_id) {
		refs = g_strconcat (Pan.active_message_id, NULL);
	}
	else {
		refs = g_strdup ( "" );
	}



	len = strlen (refs);
	if (len <= GNKSA_LENGTH) {
		return refs;
	}

	/* The reference header is too long so start truncating */
	/* We have to preserve the first message-id */
	p = strchr(refs, '>');
	if (p == NULL)
		return NULL;
	p++;
	len = p - refs;
	if (len >= GNKSA_LENGTH) {
		return refs;
	}
	strncpy(buf, refs, len);
	buf[len] = 0;
	new_refs = g_strconcat(buf, " ", NULL);
	total_len = len + 1;

	refs += len;
	if ((*refs == '\t') || (*refs == ' '))
		refs++;					/* already handled space */

	/* Now use message-ids at end of the line */
	p = refs + strlen(refs);

	while (1) {
		if (*p == '<') {
			if (strlen(p) + total_len > GNKSA_LENGTH) {
				p = strchr(p + 1, '<');
				if (p == NULL)
					p = refs;
				break;
			}
		}
		if (p == refs)
			break;
		p--;
	}

	pch = new_refs;
	new_refs = g_strconcat (new_refs, p, NULL);
	g_free ( pch );
	g_free ( refs );
	return new_refs;
}

#if 0
static gchar*
generate_message_id (void)
{
	unsigned long now;
	unsigned long randy;
	static unsigned char baseid[64];
	gchar *b, *t, tmp[32];
	char *chars32 = "0123456789abcdefghijklmnopqrstuv";
	static unsigned long last_now;

	while (1) {
		time((time_t *) & now);
		if (now != last_now)
			break;
		sleep (1);
	}
	last_now = now;
	srand (now + getpid());
	randy = (unsigned long) rand();

	b = baseid;
	t = tmp;
	while (now) {
		*t++ = chars32[now & 0x1F];
		now = now >> 5;
	}

	while (t > tmp) {
		t--;
		*b++ = *t;
	}
	*b++ = '.';

	t = tmp;
	while (randy) {
		*t++ = chars32[randy & 0x1F];
		randy = randy >> 5;
	}
	while (t > tmp) {
		t--;
		*b++ = *t;
	}

	*b++ = '@';
	return (char *) baseid;
}
#endif


/* Returns the computers host name */
static const gchar*
get_host_name (void)
{
	gchar *ptr;
	static gchar hostname[256];

	hostname[0] = '\0';

	if ((gethostname(hostname, sizeof(hostname))) == -1) {
		if ((ptr = getenv("HOST")) != NULL)
			strncpy(hostname, ptr, MAXHOSTNAMELEN);
		else if ((ptr = getenv("HOSTNAME")) != NULL)
			strncpy(hostname, ptr, MAXHOSTNAMELEN);
		else
			hostname[0] = '\0';
	}
	hostname[MAXHOSTNAMELEN] = '\0';
	ptr = strtok (hostname, ".");

	return g_strdup ( ptr );
}


/* Returns the fully qualified domain name */
static const gchar *
get_fqdn (const gchar *host)
{
	static char name[512];
	char line[1025];
	char *cp, *domain;
	static char fqdn[1024];
	struct hostent *hp;
	struct in_addr in;
	FILE *inf;

	*fqdn = '\0';
	domain = NULL;

	name[MAXHOSTNAMELEN] = '\0';
	if (host) {
		if (strchr(host, '.'))
			return host;
		strncpy(name, host, MAXHOSTNAMELEN);
	}
	else {
		if (gethostname(name, MAXHOSTNAMELEN))
			return NULL;
	}

	if ('0' <= *name && *name <= '9') {
		in.s_addr = inet_addr(name);
		if ((hp = gethostbyaddr((char *) &in.s_addr, 4, AF_INET)))
			in.s_addr = (*hp->h_addr);
		return (hp && strchr(hp->h_name, '.') ? hp->h_name : (gchar *) inet_ntoa(in));
	}
	if ((hp = gethostbyname(name)) && !strchr(hp->h_name, '.'))
		if ((hp = gethostbyaddr(hp->h_addr, hp->h_length, hp->h_addrtype)))
			in.s_addr = (*hp->h_addr);

	sprintf(fqdn, "%s", hp ? strchr(hp->h_name, '.')
		? hp->h_name : (gchar *) inet_ntoa(in)
		: "");
	if (!*fqdn || (fqdn[strlen(fqdn) - 1] <= '9')) {
		*fqdn = '\0';
		inf = fopen("/etc/resolv.conf", "r");
		if (inf) {
			while (fgets(line, 1024, inf)) {
				line[1024] = '\0';
				g_strchug(line);
				g_strchomp(line);
				if (strncmp(line, "domain ", 7) == 0) {
					domain = line + 7;
					break;
				}
				if (strncmp(line, "search ", 7) == 0) {
					domain = line + 7;
					cp = strchr(domain, ' ');
					if (cp)
						*cp = '\0';
					break;
				}
			}
			if (domain)
				sprintf(fqdn, "%s.%s", name, domain);
			fclose(inf);
		}
	}
	return (fqdn);
}

#if 0
static gchar *
reformat_body (gchar *body)
{
	gchar *tmp;
	gchar *body2 = g_malloc0 (sizeof (body));

	while ((tmp = strtok (body, "\n"))) {
		tmp = wrap_long_lines (tmp, 78);
		g_strconcat (body2, tmp, NULL);
	}

	return body2;
}
#endif

#if 0
/* Plan for checking e-mail To syntax:
 * 1. make sure there is a username (anything before the @)
 * 2. lookup hostname
 */
static gint
check_email_rcpts_syntax (GSList *rcpts)
{
	return 0;
}
#endif

#if 0
/* Plan for checking NNTP To syntax;
 * 1. check to see if it's in Pan.grouplist
 *
 */
static gint
check_nntp_rcpts_syntax (GSList *rcpts)
{
	return 0;
}
#endif


/* Find and return a list of e-mail addresses to send to */
static GSList *
get_email_rcpts (gchar *to)
{
	gchar *dest;
	GSList *email_rcpts = NULL;

	g_return_val_if_fail (to!=NULL, NULL);

	dest = strtok (to, ",");

	while (dest != NULL)
	{
		while (*dest == ' ') dest++; /* skip over whitespaces */

		if (strchr (dest, '@'))
			email_rcpts = g_slist_append ( email_rcpts, dest );

		dest = strtok (NULL, ",");
	}

	return email_rcpts;
}


/* Find and return a list of NNTP groups to send to */
static GSList *
get_nntp_rcpts (gchar *to)
{
	gchar *dest;
	GSList *nntp_rcpts = NULL;
	group_data *tmp;

	g_return_val_if_fail (to!=NULL, NULL);

	tmp = g_malloc0 (sizeof (group_data));

	dest = strtok (to, ",");

	while (dest != NULL)
	{
		while (*dest == ' ') dest++; /* skip over whitespaces */

		if ( strchr (dest, '.') && !strchr (dest, '@') ) {
			tmp->name = dest;
			nntp_rcpts = g_slist_append ( nntp_rcpts, dest );
		}

		dest = strtok (NULL, ",");
	}

	g_free (tmp);

	return nntp_rcpts;
}


static gchar*
build_nntp_message (Message *msg)
{
	gchar *subject = NULL;
	gchar *from = NULL;
	gchar *newsgroups = NULL;
	gchar *message = NULL;
	gchar *path = NULL;
	//gchar *organization = NULL;
	gchar *body = NULL;
	gchar *buf = NULL;
	static const char *host = NULL;
	static const char *fqdn = NULL;
	static char *uname = NULL;
	//gchar *err;
	GSList *l = NULL;
	gchar *tmp = NULL;

	g_return_val_if_fail (msg!=NULL, NULL);

	if (host == NULL)
		host = get_host_name ();
	if (fqdn == NULL)
		fqdn = get_fqdn (host);
	if (uname == NULL)
		uname = g_get_user_name ();

	path = g_strconcat (host, "!", uname, NULL);
	//g_free (host);

	l = msg->rcpts;
	while (l) {
		if (newsgroups)
			newsgroups = g_strconcat (newsgroups, (gchar*)l->data, NULL);
		else
			newsgroups = g_strconcat ((gchar*)l->data, NULL);

		l = l->next;
		if (l)
			newsgroups = g_strconcat (newsgroups, ",", NULL);
	}

//	newsgroups = gtk_entry_get_text (GTK_ENTRY (msg->window->newsgroups));

	from = gtk_editable_get_chars (GTK_EDITABLE (msg->window->from), 0, -1);
	subject = gtk_editable_get_chars (GTK_EDITABLE (msg->window->subject), 0, -1);
	//organization = gtk_editable_get_chars (GTK_EDITABLE (msg->window->organization), 0, -1);
	body = gtk_editable_get_chars (GTK_EDITABLE (msg->window->body), 0, -1);

#if 0
	err = check_body ( body );
	if ( err != NULL ) {
		pan_error_dialog ( _("Unable to post message:\n%s"), err );
		g_free ( body );
		return NULL;
	}
	g_free ( body );
#endif

/* FIXME: doesn't do anything
	if (!check_from_syntax(from))
		g_error ("Bad From. Do something!");
	if (!check_newsgroups_syntax(newsgroups))
		g_error ("Bad Newsgroups. Do something!");
*/
	message = g_strconcat ("From: ", from, "\n",
			       "Path: ", path, "\n",
			       "Newsgroups: ", newsgroups, "\n",
			       "Subject: ", subject, "\n", NULL);

	g_free (path);
	g_free (from);
	g_free (newsgroups);
	g_free (subject);
	tmp = message;

	/* FIXME
	if (Pan.active_server->gen_msgid) {
		message = g_strconcat (tmp,"Message-ID: <", generate_message_id(), fqdn, ">\n", NULL);
		g_free (tmp);
		tmp = message;
	}
	*/
	if (msg) {
		gchar* pch = generate_references ( msg );
		message = g_strconcat (message, "References: ", pch, "\n", NULL );
		g_free ( pch );
		g_free ( tmp );
		tmp = message;
	}
	body = g_strconcat ("\n",
		gtk_editable_get_chars(GTK_EDITABLE(msg->window->body), 0, -1),
		NULL);
	buf = generate_date ();
	message = g_strconcat (tmp,
			       "Date: ", buf, "\n",
			       //"Organization: ", organization, "\n",
			       "X-Newsreader: ", "Pan ", VERSION, "\n",
			       body, NULL);
	g_free (buf);
	g_free (tmp);
	g_free (body);

	return message;
}

static gchar*
check_subject (const gchar *subject, GoodnessLevel* goodness)
{
	gchar *pch = NULL;
	gchar *tmp = NULL;

	g_assert (goodness!=NULL);

	if (!subject || !*subject) {
		tmp = g_strjoin ( "\n", _("No Subject Specified."), pch, NULL );
		g_free (pch);
		pch = tmp;
		if (*goodness<REFUSE)
			*goodness = REFUSE;
	}

	return pch;
}


typedef struct
{
	Message *message;
	GtkWidget *send_button;
	gint status;
}
SendDoneStruct;

static int
send_done_cb_mainthread (gpointer data)
{
	SendDoneStruct *s = (SendDoneStruct*) data;
	gchar *msg = NULL;
	gchar *tmp = NULL;
	g_message ( "send_done_cb_mainthread %p", data );

	if (s->status & SMTP_FAILED) {
		tmp = g_strjoin ("\n", _("E-Mail send failed."), msg, NULL);
		g_free (msg);
		msg = tmp;
	}
	if (s->status & NNTP_FAILED) {
		tmp = g_strjoin ("\n", _("Usnet posting failed."), msg, NULL);
		g_free (msg);
		msg = tmp;
	}
	if (!s->status) {
		tmp = g_strjoin ("\n", _("Message sent successfully."), msg, NULL);
		g_free (msg);
		msg = tmp;
	}

	gnome_ok_dialog_parented (msg, GTK_WINDOW(s->message->window->window));

	if (!s->status) {
		gtk_widget_destroy (s->message->window->window);
		message_free (s->message);
	} else {
		gtk_widget_set_sensitive (s->send_button, TRUE);
	}

	g_free (s);

	return 0;
}

static void
send_done_cb (StatusItem* item, int status)
{
	SendDoneStruct *s = g_new0 (SendDoneStruct, 1);
	s->message = QUEUE_ITEM_SEND(item)->msg;
	s->send_button = GTK_WIDGET(gtk_object_get_data(GTK_OBJECT(item),"send_button"));
	s->status = status;
	gui_invoke_in_gui_thread (send_done_cb_mainthread, s);
}

#if 0
typedef struct
{
	GtkWidget *w;
	Message *msg;
}
AreYouSureStruct;

void
are_you_sure_you_want_to_post_cb (gint reply, gpointer data)
{
	AreYouSureStruct *sure = (AreYouSureStruct*) data;
	GtkWidget *w = sure->w;
	Message *msg = sure->msg;
	g_free (sure);
	gtk_

	if (data==0) /* yes */
#endif




void
send_cb (GtkWidget *widget, Message *msg)
{
	GtkObject *item = NULL;
	gchar* pch = NULL;
	gchar *subject = NULL;
	gchar *body = NULL;
	GoodnessLevel goodness = OKAY;
	GSList *warnings = NULL;

	g_return_if_fail (widget!=NULL);
	g_return_if_fail (msg!=NULL);

	g_message ( "send_cb" );
	g_message ( "checking subject..." );

	/* check the subject... */
	subject = g_strdup(gtk_entry_get_text(GTK_ENTRY(msg->window->subject)));
	subject = g_strstrip(subject);
	pch = check_subject (subject, &goodness);
	g_free (subject);
	if (pch!=NULL) {
		g_message (pch);
		if (goodness==REFUSE)
			pan_error_dialog (pch);
		else if (goodness==WARN)
			warnings = g_slist_append(warnings,g_strdup(pch));
		g_free (pch);
		if (goodness==REFUSE)
			return;
	}

	g_message ( "checking body..." );
	/* check the body... */
	body = gtk_editable_get_chars (GTK_EDITABLE (msg->window->body), 0, -1);
	body = g_strchug (g_strchomp(body));
	pch = check_body (body, &goodness);
	g_free (body);
	if (pch!=NULL) {
		g_message (pch);
		if (goodness==REFUSE)
			pan_error_dialog (pch);
		else if (goodness==WARN)
			warnings = g_slist_append(warnings,g_strdup(pch));
		g_free (pch);
		if (goodness==REFUSE)
			return;
	}

	/* FIXME: check to: and from: fields */


	if (warnings!=NULL)
		g_warning ("FIXME: warnings");

	/* only queue it after it has passed all the checks. */

	item = queue_item_send_new (grouplist_get_current_server(),msg);
	gtk_object_set_data (item, "send_button", widget);
	gtk_widget_set_sensitive (GTK_WIDGET(widget), FALSE);
	gtk_signal_connect (item, "done", send_done_cb, NULL);
	queue_add (QUEUE_ITEM(item));
}

gint
message_send (StatusItem *item, PanSocket *sock, Message *msg)
{
	GSList *nntp_rcpts;
	GSList *email_rcpts;
	gchar *nntp_msg;
	gchar *tmp1, *tmp2;
	gchar *newsgroups;
	gchar *users;
	gchar *subject;
	gint retval = 0;

	g_return_val_if_fail (msg!=NULL, ~0);

	subject = gtk_editable_get_chars (
		GTK_EDITABLE (msg->window->subject), 0, -1);

	/* where is this being posted... */
	newsgroups = gtk_editable_get_chars (
		GTK_EDITABLE (msg->window->newsgroups), 0, -1);
	tmp1 = g_strdup (newsgroups);
	nntp_rcpts = get_nntp_rcpts (tmp1);

	/* who is this being mailed to... */
	users = gtk_editable_get_chars (
		GTK_EDITABLE (msg->window->users), 0, -1);
	tmp2 = g_strdup (users);
	email_rcpts = get_email_rcpts (tmp2);


	if (email_rcpts) {
		msg->rcpts = email_rcpts;
		if (!smtp_send (item, msg))
			log_add_va (_("E-mail Message '%s' sent."), subject);
		else {
			log_add_va (_("E-mail Message '%s' not sent."),subject);
			retval |= SMTP_FAILED;
		}
	}


	if (nntp_rcpts) {
		msg->rcpts = nntp_rcpts;
		nntp_msg = build_nntp_message (msg);

		if (!nntp_post (sock, nntp_msg))
			log_add_va (_("NNTP Article '%s' posted."), subject);
		else
			retval |= NNTP_FAILED;

		g_free (nntp_msg);
	}


	/* clean up */
	g_free (subject);
	g_free (tmp1);
	g_free (tmp2);
	g_free (newsgroups);
	g_free (users);

	return retval;
}
