/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <stdlib.h>
#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "acache.h"
#include "article.h"
#include "article-db.h"
#include "debug.h"
#include "log.h"
#include "nntp.h"
#include "queue.h"
#include "queue-item-body.h"
#include "text.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static void queue_item_body_class_init (QueueItemBodyClass *klass);

static void queue_item_body_init (QueueItemBody *item);

static gint queue_item_body_run (QueueItem* item);

static void queue_item_body_destroy (GtkObject* item);

static gchar* queue_item_body_describe (StatusItem* item);

static int nntp_article_download (QueueItem*, article_data*);


/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

extern gboolean must_exit;

/***********
************  Public
***********/

/***********
************  Private
***********/

static QueueItemClass *parent_class = NULL;

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

GtkType
queue_item_body_get_type (void)
{
	static GtkType queue_item_body_type = 0;

	if (!queue_item_body_type)
	{
		static const GtkTypeInfo queue_item_body_info =
		{
			"QueueItemBody",
			sizeof(QueueItemBody),
			sizeof(QueueItemBodyClass),
			(GtkClassInitFunc) queue_item_body_class_init,
			(GtkObjectInitFunc) queue_item_body_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};

		queue_item_body_type = gtk_type_unique (
			TYPE_QUEUE_ITEM,
			&queue_item_body_info);
	}

	return queue_item_body_type;
}

static void
queue_item_body_class_init (QueueItemBodyClass *klass)
{
        parent_class = gtk_type_class (TYPE_QUEUE_ITEM);

	GTK_OBJECT_CLASS(klass)->destroy = queue_item_body_destroy;
	STATUS_ITEM_CLASS(klass)->describe = queue_item_body_describe;
	QUEUE_ITEM_CLASS(klass)->run = queue_item_body_run;
}

static void
queue_item_body_init (QueueItemBody *item)
{
	item->gdata = NULL;
	item->sdata = NULL;
	item->message = NULL;
}

/*****
******
*****/

GtkObject*
queue_item_body_new (server_data *sdata,
                     group_data *gdata,
		     Message *msg)
{
	QueueItemBody *item = NULL;

	/* sanity checks */
	g_return_val_if_fail(sdata!=NULL, NULL);
	g_return_val_if_fail(gdata!=NULL, NULL);
	g_return_val_if_fail(msg!=NULL, NULL);
	g_return_val_if_fail(msg->message_id!=NULL, NULL);
	g_return_val_if_fail(*msg->message_id!='\0', NULL);

	/* gtk instantiation */
	item = QUEUE_ITEM_BODY(gtk_type_new(TYPE_QUEUE_ITEM_BODY));

	/* construct superclass... */
	queue_item_construct (QUEUE_ITEM(item), sdata, TRUE, TRUE);

	/* construct this class... */
	item->gdata = gdata;
	item->sdata = sdata;
	item->message = msg;

	return GTK_OBJECT(item);
}

/*****
******
*****/

static gchar*
queue_item_body_describe (StatusItem* item)
{
	QueueItemBody* body = QUEUE_ITEM_BODY(item);

	/* sanity checks */
        g_return_val_if_fail (item!=NULL, NULL);
	g_return_val_if_fail (IS_QUEUE_ITEM_BODY(item), NULL);
	g_return_val_if_fail (GTK_OBJECT_CONSTRUCTED(item), NULL);
	g_return_val_if_fail (body->message!=NULL, NULL);
	g_return_val_if_fail (body->message->message_id!=NULL, NULL);
	g_return_val_if_fail (*body->message->message_id!='\0', NULL);
	g_return_val_if_fail (body->gdata->name!=NULL, NULL);
	g_return_val_if_fail (*body->gdata->name!='\0', NULL);

	return g_strdup_printf ("Downloading body of ``%s'' from group %s",
			QUEUE_ITEM_BODY(item)->message->message_id,
			QUEUE_ITEM_BODY(item)->gdata->name );
}

/*****
******
*****/

static void
queue_item_body_destroy (GtkObject* object)
{
        debug (DEBUG_QUEUE,
		"GtkObject '%p' in queue_item_body_destroy", object);

	/* sanity checks */
        g_return_if_fail (object != NULL);
	g_return_if_fail (IS_QUEUE_ITEM_BODY (object));
	g_return_if_fail (GTK_OBJECT_CONSTRUCTED (object));

	/* clean up this class */
	/* (nothing -- text_update() frees the Message) */

	/* invoke superclass */
	if (GTK_OBJECT_CLASS(parent_class)->destroy)
		(*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}
/*****
******
*****/

static gint
queue_item_body_run (QueueItem* item)
{
	server_data *sdata = item->sdata;
	group_data *gdata = QUEUE_ITEM_BODY(item)->gdata;
	Message *msg = QUEUE_ITEM_BODY(item)->message;
	article_db db = NULL;
	article_data *adata = NULL;

	/* sanity checks */
	g_return_val_if_fail(sdata!=NULL, -1);
	g_return_val_if_fail(gdata!=NULL, -1);
	g_return_val_if_fail(msg!=NULL, -1);
	g_return_val_if_fail(msg->message_id!=NULL, -1);
	g_return_val_if_fail(*msg->message_id!='\0', -1);

	/* get the article header */
	db = ahdb_ref (sdata, gdata);
	adata = ahdb_get ( db, (const char*)msg->message_id );
	ahdb_unref ( db );

	/* make sure we've got the first header */
	if ( !adata ) {
		status_item_emit_error(STATUS_ITEM(item),
			_("Couldn't find article header for message_id '%s'"),
			msg->message_id );
		return -1;
	}

	article_remove_flag (sdata, gdata, adata, STATE_DOWNLOAD_QUEUED, TRUE);

	/* Check the cache before we download */
	if (!acache_header_exists(adata->message_id) ||
            !acache_body_exists(adata->message_id))
	{
		if ( nntp_article_download(item,adata) ) {
			status_item_emit_status(STATUS_ITEM(item),
				_("Downloading article ``%s'' failed."),
				adata->subject);
			article_free (adata);
			return -1;
		}
	}

	/* Display, et cetera */
	if (item->high_priority)
	{
		message_import_adata (msg, adata);
		if (adata->state != STATE_READ) {
			group_add_attrib_i (sdata, gdata, "Read", 1);
			article_add_flag (sdata, gdata, adata, STATE_READ, TRUE);
		}
		text_update (msg);
	}

	article_free (adata);
	return 0;
}

/*****
******
*****/

static int
nntp_article_download (QueueItem *item, article_data* adata)
{
	PanSocket *sock = item->sock;
	group_data *gdata = QUEUE_ITEM_BODY(item)->gdata;
	server_data *sdata = QUEUE_ITEM_BODY(item)->sdata;
	gboolean adata_headers_changed = FALSE;
	const char *buffer;
	FILE *f;

	/* sanity checks */
	g_return_val_if_fail (sock!=NULL, -1);

	status_item_emit_status (STATUS_ITEM(item),
		_("Downloading article '%s'"), adata->subject);

	/* set group */
	if (nntp_set_group(sock, gdata)) {
		status_item_emit_error (STATUS_ITEM(item),
			_("Couldn't select group '%s'"), gdata->name);
		return -1;
	}

	/* get the head */
	if (pan_socket_putline_va (sock, "HEAD %s\r\n", adata->number)) {
		status_item_emit_sockwrite_err(STATUS_ITEM(item));
		return -1;
	}
	if (pan_socket_getline (sock, &buffer)) {
		status_item_emit_sockread_err(STATUS_ITEM(item));
		return -1;
	}
	if (strtol(buffer, (char **) NULL, 10) != 221) {
		status_item_emit_error(STATUS_ITEM(item),
			_("NNTP \"HEAD\" command returned %d.\n"
			  "The article may have expired."),
			  buffer);
		return -1;
	}
	if (pan_socket_getline (sock, &buffer)) {
		status_item_emit_sockread_err(STATUS_ITEM(item));
		return -1;
	}
	
	f = acache_open_header (adata, ACACHE_NONE);
	while (strncmp(buffer, ".\r\n", 3))
	{
		char* pch = NULL;
		const char* name = NULL;

		/* check for exit... */
		if (must_exit) {
			acache_close_header (f);
			acache_delete (adata->message_id);
			return -1;
		}

		/* write to acache... */
		acache_putline_header (f, buffer);

		/* parse it for adata... */
		if ((pch = strrchr (buffer, '\x0D'))) /* trim linefeed */
			*pch = '\0';
		name = "Newsgroups: ";
		if (!strncmp ( buffer, name, strlen(name))) {
			adata->newsgroups = g_strdup ( buffer+strlen(name) );
			adata_headers_changed = TRUE;
		}
		name = "Reply-To: ";
		if (!strncmp (buffer, name, strlen(name))) {
			adata->reply_to = g_strdup ( buffer+strlen(name) );
			adata_headers_changed = TRUE;
		}
		name = "Followup-To: ";
		if (!strncmp ( buffer, name, strlen(name))) {
			adata->followup_to = g_strdup ( buffer+strlen(name) );
			adata_headers_changed = TRUE;
		}

		/* get the next line... */
		if (pan_socket_getline (sock, &buffer)) {
			status_item_emit_sockread_err(STATUS_ITEM(item));
			acache_close_header (f);
			acache_delete (adata->message_id);
			return -1;
		}
	}
	acache_close_header (f);

	if (adata_headers_changed)
		article_update (sdata, gdata, adata, TRUE);

	if (pan_socket_putline_va (sock, "BODY %s\r\n", adata->number)) {
		status_item_emit_sockwrite_err(STATUS_ITEM(item));
		return -1;
	}
	if (pan_socket_getline (sock, &buffer)) {
		status_item_emit_sockread_err(STATUS_ITEM(item));
		return -1;
	}
	if (strtol(buffer, (char **) NULL, 10) != 222) {
		return -1;
	}
	
	if (pan_socket_getline (sock, &buffer)) {
		status_item_emit_sockread_err(STATUS_ITEM(item));
		return -1;
	}
	f = acache_open_body (adata, ACACHE_NONE);

	status_item_emit_init_steps (STATUS_ITEM(item), adata->linecount);

	while (strncmp(buffer, ".\r\n", 3))
	{
		if (must_exit)
		{
			acache_close_body (f);
			acache_delete (adata->message_id);
			return -1;
		}
		acache_putline_body (f, buffer);
		status_item_emit_next_step(STATUS_ITEM(item));
		if (pan_socket_getline (sock, &buffer))
		{
			status_item_emit_sockread_err(STATUS_ITEM(item));
			acache_close_body (f);
			acache_delete (adata->message_id);
			return -1;
		}
	}
	acache_close_body (f);
	status_item_emit_progress(STATUS_ITEM(item), 100);

	return 0;
}
