/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <string.h>

#include <glib.h>
#include <libgnome/gnome-defs.h>
#include <libgnome/gnome-i18n.h>

#include "debug.h"
#include "queue-item.h"
#include "status-item.h"

/*********************
**********************  Defines / Enumerated types
*********************/

enum {
	RUN,
	LAST_SIGNAL
};

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static void queue_item_class_init (QueueItemClass *klass);

static void queue_item_init (QueueItem *sock);

static void queue_item_destroy (GtkObject* object);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

static guint queue_item_signals[LAST_SIGNAL] = { 0 };

static GtkObjectClass *parent_class = NULL;


/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

GtkType
queue_item_get_type (void)
{
	static GtkType queue_item_type = 0;

	if (!queue_item_type)
	{
		static const GtkTypeInfo queue_item_info =
		{
			"QueueItem",
			sizeof(QueueItem),
			sizeof(QueueItemClass),
			(GtkClassInitFunc) queue_item_class_init,
			(GtkObjectInitFunc) queue_item_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};

		queue_item_type = gtk_type_unique (
			TYPE_STATUS_ITEM,
			&queue_item_info);
	}

	return queue_item_type;
}

/*****
******
*****/

typedef int (*GtkSignal_INT__NONE)(GtkObject*, gpointer);

static void marshal_INT__NONE(
	GtkObject *obj, GtkSignalFunc func, gpointer func_data, GtkArg *args)
{
	GtkSignal_INT__NONE rfunc = (GtkSignal_INT__NONE) func;
	gint *retval = GTK_RETLOC_INT(args[0]);
	*retval = (*rfunc)(obj, func_data);
}

static void
queue_item_class_init (QueueItemClass *klass)
{
	GtkObjectClass *object_class = GTK_OBJECT_CLASS(klass);
	StatusItemClass *status_class = STATUS_ITEM_CLASS(klass);

	parent_class = gtk_type_class (TYPE_STATUS_ITEM);

	queue_item_signals[RUN] =
		gtk_signal_new ( "run",
			GTK_RUN_LAST,
			object_class->type,
			GTK_SIGNAL_OFFSET(QueueItemClass,run),
			marshal_INT__NONE,
			GTK_TYPE_INT, 0 );

        gtk_object_class_add_signals (
		object_class, 
		queue_item_signals,
		LAST_SIGNAL);

	object_class->destroy = queue_item_destroy;
	status_class->describe = NULL; /* pure virtual function */
	klass->run = NULL; /* pure virtual function */
}

static void
queue_item_init (QueueItem *item)
{
	item->sdata = NULL;
	item->high_priority = FALSE;
	item->needs_socket = TRUE;
	memset ( &item->thread_id, 0, sizeof(pthread_t) );
}

/*****
******
*****/

void
queue_item_construct (QueueItem* item,
                      server_data *sdata,
	              gboolean high_priority,
		      gboolean needs_socket)
{
	g_return_if_fail (item!=NULL);
	g_return_if_fail (IS_QUEUE_ITEM(item));

	item->sdata = sdata;
	item->high_priority = high_priority;
	item->needs_socket = needs_socket;
}

static void
queue_item_destroy (GtkObject* object)
{
	debug(DEBUG_QUEUE, "GtkObject '%p' in queue_item_destroy", object);

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_QUEUE_ITEM (object));
	g_return_if_fail (GTK_OBJECT_CONSTRUCTED (object));

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

/*****
******
*****/

gint
queue_item_run (QueueItem *item)
{
	gint retval = FALSE;

	g_return_val_if_fail (item!=NULL, -1);
	g_return_val_if_fail (IS_QUEUE_ITEM(item), -1);

	gtk_signal_emit (GTK_OBJECT(item), queue_item_signals[RUN], &retval);

	return retval;
}

/*****
******
*****/

gboolean
queue_item_is_high_priority (const QueueItem *item)
{
	g_return_val_if_fail (item!=NULL, FALSE);
	g_return_val_if_fail (IS_QUEUE_ITEM(item), FALSE);

	return item->high_priority;
}

/*****
******
*****/

gboolean
queue_item_needs_socket (const QueueItem *item)
{
	g_return_val_if_fail (item!=NULL, NULL);
	g_return_val_if_fail (IS_QUEUE_ITEM(item), NULL);

	return item->needs_socket;
}

PanSocket*
queue_item_get_socket (QueueItem *item)
{
	g_return_val_if_fail (item!=NULL, NULL);
	g_return_val_if_fail (IS_QUEUE_ITEM(item), NULL);

	return item->sock;
}

void
queue_item_set_socket (QueueItem *item, PanSocket *sock)
{
	g_return_if_fail (item!=NULL);
	g_return_if_fail (IS_QUEUE_ITEM(item));
	g_return_if_fail (sock!=NULL);

	item->sock = sock;
}

