/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/types.h>
#include <fcntl.h>
#include <pthread.h>

#include "debug.h"
#include "sockets.h"

/*********************
**********************  Defines / Enumerated types
*********************/

#define BUFSIZE 8192

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static void pan_socket_class_init (PanSocketClass *klass);

static void pan_socket_init (PanSocket *sock);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

extern int h_errno;

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

GtkType
pan_socket_get_type (void)
{
	static GtkType socket_type = 0;

	if (!socket_type)
	{
		static const GtkTypeInfo socket_info =
		{
			"PanSocket",
			sizeof(PanSocket),
			sizeof(PanSocketClass),
			(GtkClassInitFunc) pan_socket_class_init,
			(GtkObjectInitFunc) pan_socket_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};

		socket_type = gtk_type_unique (GTK_TYPE_OBJECT, &socket_info);
	}

	return socket_type;
}

static void
pan_socket_class_init (PanSocketClass *class)
{
}

static void
pan_socket_init (PanSocket *sock)
{
	sock->sockfd = 0;
	sock->inbuf = NULL;
	sock->endbuf = NULL;
	sock->inptr = NULL;
	sock->line_buffer_size = 0;
	sock->line_buffer = NULL;
	sock->last_action_time = 0;
}

/*****
******
*****/

GtkObject*
pan_socket_new (const gchar* address,
                gint port)
{
	PanSocket *sock = NULL;
	struct sockaddr_in serv_addr;
	struct hostent host_info;
	int on;
	unsigned long inaddr;
	struct hostent *hp;
	int sockfd;
#ifdef TRY_AGAIN
	unsigned int max_retries = 3;
#endif

	/* sanity checks */
	g_return_val_if_fail (address!=NULL, NULL);
	g_return_val_if_fail (port>=0, NULL);

	/* set up the serv_addr struct */
	memset ( &serv_addr, 0, sizeof(serv_addr) );
	serv_addr.sin_family = AF_INET;
	serv_addr.sin_port = htons(port);

	/* Try to convert host name as dotted decimal */
	inaddr = inet_addr(address);
	if ( inaddr != -1 )
	{
		memcpy(&serv_addr.sin_addr,&inaddr,sizeof(inaddr));
		host_info.h_name = NULL;
	}
	else /* If that failed, then look up the host name */
	{
		while (NULL==(hp = gethostbyname(address)))
		{
#ifdef TRY_AGAIN
			max_retries--;
			if (max_retries && (h_errno == TRY_AGAIN)) {
				sleep(1);
				continue;
			}
#endif
			g_warning (
				"Can't resolve %s -- PanSocket not created",
			       	address);
			return NULL;
		}
		host_info = *hp;
		memcpy(&serv_addr.sin_addr, hp->h_addr, hp->h_length);
	}

	sockfd = socket(AF_INET, SOCK_STREAM, 0);
	if (sockfd < 0) {
		g_warning ("Can't create socket: %s", strerror(errno));
		return NULL;
	}

	if (connect(sockfd, (struct sockaddr*)&serv_addr, sizeof(serv_addr))<0)
	{
		close(sockfd);
		g_warning ("socket connect failed: %s", strerror(errno));
		return NULL;
	}

	on = 1;
	if (setsockopt (sockfd, SOL_SOCKET, SO_KEEPALIVE,
	                (char*)&on, sizeof(on)) < 0)
	{
		close(sockfd);
		g_warning ("socket keepalive option failed: %s",
			strerror(errno));
		return NULL;
	}

	sock = gtk_type_new ( PAN_TYPE_SOCKET );

	sock->sockfd = sockfd;
	sock->inbuf = g_malloc0 ( BUFSIZE );
	sock->endbuf = sock->inbuf;
	sock->inptr = sock->inbuf;
	sock->line_buffer_size = 512;
	sock->line_buffer = g_malloc0 ( sock->line_buffer_size );

	debug(DEBUG_SOCKET_OUTPUT|DEBUG_SOCKET_INPUT,
		"socket created: %p for %s, port %d", sock, address, port );
	return GTK_OBJECT(sock);
}


void
pan_socket_destroy (PanSocket *sock)
{
	g_return_if_fail (sock != NULL);
	g_return_if_fail (PAN_IS_SOCKET (sock));
	g_return_if_fail (GTK_OBJECT_CONSTRUCTED (sock));

	debug(DEBUG_SOCKET_OUTPUT|DEBUG_SOCKET_INPUT,
		"socket closed: %p", sock);

	/* free up our pieces */
	close (sock->sockfd);
	g_free (sock->line_buffer);
	g_free (sock->inbuf);

	/* pass along to superclass */
	gtk_object_destroy ((GtkObject*) sock);
}

/*****
******
*****/

/* simple read buffering to hopefully speed things up. */
static int
read_char (PanSocket* ps, gchar* c)
{
	if ( ps->inptr == ps->endbuf )
	{
		int read_count = 0;
		int retval = 0;
		struct timeval tv;
		fd_set rfds;
		FD_ZERO (&rfds);
		FD_SET (ps->sockfd, &rfds);
		/* We will timeout after 2 minutes. */
		tv.tv_sec = 120;
		tv.tv_usec = 0;
		retval = select (ps->sockfd + 1, &rfds, NULL, NULL, &tv);
		if (retval <= 0) {
			g_warning ("Timed out waiting for select()");
			return -1;
		}
		read_count = read (ps->sockfd, ps->inbuf, BUFSIZE);
		ps->endbuf = ps->inbuf + read_count;
		ps->inptr = ps->inbuf;
		if ( read_count <= 0 )
		   return read_count;
	}

	*c = *ps->inptr++;

	return 1;
}


/*
 * get_server - read from a socket until \n
 */
const gint
pan_socket_getline(PanSocket *sock, const gchar **setme)
{
	gchar c;
	int len = 0;

	g_assert ( sock->line_buffer_size>0 );

	sock->last_action_time=time(0);

	for ( ;; )
	{
		/* get the next char */
		if ( read_char(sock, &c) != 1 ) {
			debug(DEBUG_SOCKET_INPUT, "socket [%p] failed its read", sock);
			return -1;
		}

		/* grow the line buffer if we need to */
		if (len >= sock->line_buffer_size - 1)
		{
			sock->line_buffer_size *= 2;
			sock->line_buffer = g_realloc (
				sock->line_buffer,
				sock->line_buffer_size );
		}

		/* add the char to our line buffer */
		sock->line_buffer[len++] = c;

		if ( c == '\n' )
			break;
	}

	sock->line_buffer[len] = '\0';
	if ( !len || sock->line_buffer[len-1]!='\n' ) {
		debug(DEBUG_SOCKET_INPUT, "socket [%p] failed its read", sock);
		return -1;
	}

	debug(DEBUG_SOCKET_INPUT,
		"socket [%p] received [%s]", sock, sock->line_buffer);

	//pan_stats.bytes_received += len;

	*setme = sock->line_buffer;
	return 0;
}


gint 
pan_socket_putline (PanSocket     *sock,
                    const gchar   *line)
{
	const gchar* pch = line;
	gint nbytes = strlen (line);
	gint nleft = nbytes;

	sock->last_action_time=time(0);

	debug(DEBUG_SOCKET_OUTPUT,
		"[%ld][socket %p putline][%s]", time(0), (void*)sock, line );

	while ( nleft>0 )
	{
		gint nwritten = write(sock->sockfd, pch, nleft);
		if (nwritten == -1)
			return -1;
		if (nwritten > 0)
			nleft -= nwritten;
		pch += nwritten;
	}

	return nleft;
}


gboolean 
pan_socket_putline_va (PanSocket    *sock,
                       const gchar  *format,
                       ...)
{
	va_list args;
	char* line;
	gint retval;

	g_return_val_if_fail (format!=NULL, -1);

	va_start(args, format);
	line = g_strdup_vprintf(format, args);
	va_end(args);

	retval = pan_socket_putline (sock, line);

	g_free (line);
	return retval;
}

/*****
******
*****/

