/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*********************
**********************  Includes
*********************/

#include <gtk/gtkvbox.h>

#include "debug.h"
#include "gui.h"
#include "status-item-view.h"
#include "util.h"

/*********************
**********************  Defines / Enumerated types
*********************/

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

/*********************
**********************  Private Function Prototypes
*********************/

static void status_item_view_class_init (StatusItemViewClass *klass);
static void status_item_view_init (StatusItemView *object);
static void status_item_view_destroy (GtkObject *object);

static void status_item_view_unregister (StatusItemView *view);

static void status_item_view_progress_init_steps (StatusItem*, gint steps, StatusItemView*);
static void status_item_view_progress_next_step (StatusItem*, StatusItemView*);
static void status_item_view_progress_set_percent (StatusItem*, gint of_100, StatusItemView*);
static void status_item_view_progress_set_activity (StatusItem*, gint n, StatusItemView*);
static void status_item_view_status (StatusItem*, const gchar*, StatusItemView*);
static void status_item_view_error (StatusItem*, const gchar*, StatusItemView*);
static void status_item_view_done (StatusItem*, gint, StatusItemView*);

static gchar* string_shorten (const gchar* original);

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

static GtkObjectClass *parent_class = NULL;

static GtkTooltips *tips = NULL;

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

GtkType
status_item_view_get_type (void)
{
	static GtkType status_item_view_type = 0;

	if (!status_item_view_type)
	{
		static const GtkTypeInfo status_item_view_info =
		{
			"StatusItemView",
			sizeof(StatusItemView),
			sizeof(StatusItemViewClass),
			(GtkClassInitFunc) status_item_view_class_init,
			(GtkObjectInitFunc) status_item_view_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL
		};

		status_item_view_type = gtk_type_unique (
			GTK_TYPE_VBOX,
			&status_item_view_info);
	}

	return status_item_view_type;
}

/*****
******
*****/

static void
status_item_view_class_init (StatusItemViewClass *klass)
{
	GtkObjectClass *object_class = (GtkObjectClass*) klass;

	parent_class = gtk_type_class (GTK_TYPE_OBJECT);

	tips = gtk_tooltips_new ();

	object_class->destroy = status_item_view_destroy;
}

static void
status_item_view_init (StatusItemView *view)
{
	view->progress_major = NULL;
	view->progress_minor = NULL;
	view->description = NULL;

	view->steps = 0;
	view->step = 0;

	view->signal_id_progress_set_percent = 0;
	view->signal_id_progress_init_steps = 0;
	view->signal_id_progress_next_step = 0;
	view->signal_id_progress_set_activity = 0;
	view->signal_id_status = 0;
	view->signal_id_error = 0;
	view->signal_id_done = 0;

	view->item = NULL;
}


/*****
******
*****/

static void
status_item_view_destroy (GtkObject* object)
{
	debug(DEBUG_QUEUE, "GtkObject '%p' in status_item_view_destroy", object);

	g_return_if_fail (object != NULL);
	g_return_if_fail (IS_STATUS_ITEM_VIEW (object));
	g_return_if_fail (GTK_OBJECT_CONSTRUCTED (object));

	status_item_view_unregister (STATUS_ITEM_VIEW(object));

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(*GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

GtkWidget*
status_item_view_new (void)
{
        /* gtk instantiation */
	StatusItemView *view = STATUS_ITEM_VIEW(
		gtk_type_new(TYPE_STATUS_ITEM_VIEW));

	/* parent setup */
	GTK_CONTAINER(view)->border_width = 0;
	GTK_BOX(view)->spacing = 0;
	GTK_BOX(view)->homogeneous = FALSE;

	/* status_item_view setup */
	view->progress_major = gtk_progress_bar_new ();
	gtk_progress_set_format_string (GTK_PROGRESS(view->progress_major), "");
	gtk_progress_set_show_text (GTK_PROGRESS(view->progress_major), TRUE);
	gtk_widget_show (view->progress_major);
	view->eventbox = gtk_event_box_new();
	gtk_widget_show (view->eventbox);
	gtk_container_add(GTK_CONTAINER(view->eventbox), view->progress_major);
	view->description = gtk_label_new ("hello world");
	//gtk_widget_show (view->description);
	gtk_box_pack_start(GTK_BOX(view), view->description, FALSE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(view), view->eventbox, FALSE, TRUE, 0);

	return GTK_WIDGET(view);
}

GtkWidget*
status_item_view_new_with_item (StatusItem* item)
{
	StatusItemView* l = STATUS_ITEM_VIEW(status_item_view_new());
	status_item_view_set_item (l, item);
	return GTK_WIDGET(l);
}

/*****
******
*****/

static void
status_item_view_unregister (StatusItemView *view)
{
	GtkObject* o = view->item==NULL ? NULL : GTK_OBJECT(view->item);
	if (o==NULL)
		return;

	/* stop listening to the item... */
	gtk_signal_disconnect(o, view->signal_id_progress_set_percent);
	gtk_signal_disconnect(o, view->signal_id_progress_init_steps);
	gtk_signal_disconnect(o, view->signal_id_progress_next_step);
	gtk_signal_disconnect(o, view->signal_id_progress_set_activity);
	gtk_signal_disconnect(o, view->signal_id_status);
	gtk_signal_disconnect(o, view->signal_id_error);
	gtk_signal_disconnect(o, view->signal_id_done);

	/* unref the item... */
	gtk_object_unref(GTK_OBJECT(view->item));
	view->item = NULL;

	/* zero out our variables.. */
	view->signal_id_progress_set_percent = 0;
	view->signal_id_progress_init_steps = 0;
	view->signal_id_progress_next_step = 0;
	view->signal_id_progress_set_activity = 0;
	view->signal_id_status = 0;
	view->signal_id_error = 0;
	view->signal_id_done = 0;
}

void
status_item_view_set_item (StatusItemView* view, StatusItem* item)
{
	gchar *description = NULL;
	gchar *desc = NULL;
	GtkObject* obj = NULL;

	/* sanity checks */
	g_return_if_fail (IS_STATUS_ITEM(item));

	/* skip the no-op */
	if (item == view->item)
		return;

	description = status_item_describe (item);
	desc = string_shorten (description);
	obj = GTK_OBJECT(item);

	status_item_view_unregister (view);

	gtk_object_ref(obj);

	gtk_tooltips_set_tip(tips,
		GTK_WIDGET(view->eventbox), description, NULL);

	gtk_label_set_text (GTK_LABEL(view->description), desc);
	g_free (description);
	g_free (desc);

	view->signal_id_progress_init_steps = gtk_signal_connect (obj,
		"progress_init_steps",
		status_item_view_progress_init_steps,
		view);
	view->signal_id_progress_next_step = gtk_signal_connect (obj,
		"progress_next_step",
		status_item_view_progress_next_step,
		view);
	view->signal_id_progress_set_percent = gtk_signal_connect (obj,
		"progress_set_percent",
		status_item_view_progress_set_percent,
		view);
	view->signal_id_progress_set_activity = gtk_signal_connect (obj,
		"progress_set_activity",
		status_item_view_progress_set_activity,
		view);
	view->signal_id_status = gtk_signal_connect (obj,
		"status",
		status_item_view_status,
		view);
	view->signal_id_done = gtk_signal_connect (obj,
		"done",
		status_item_view_done,
		view);
	view->signal_id_error = gtk_signal_connect (obj,
		"error",
		status_item_view_error,
		view);

	view->item = item;
}

/*****
******
*****/

static gint
progress_set_percent (gpointer data)
{
	/* setup */
	StatusItemView *view = STATUS_ITEM_VIEW(data);
	GtkProgress *prog = GTK_PROGRESS(view->progress_major);

	/* make sure we're in percentage mode */
	if (prog->activity_mode)
		gtk_progress_set_activity_mode (prog, FALSE);

	/* update the percentage */
	view->progress_last_value = view->progress_value;
	gtk_progress_set_percentage (prog, view->progress_value/100.0);

	return 0;
}
	
static void
status_item_view_progress_set_percent (StatusItem* item,
                                      gint of_100,
				      StatusItemView* view)
{
	view->progress_value = of_100;

	g_assert (IS_STATUS_ITEM(item));
	g_assert (IS_STATUS_ITEM_VIEW(view));

	if (view->progress_last_value != view->progress_value)
		gui_invoke_in_gui_thread (progress_set_percent, view);
}


static void
status_item_view_progress_init_steps (
	StatusItem* item,
	gint steps,
	StatusItemView* view)
{
	view->step = 0;
	view->steps = steps;
	view->progress_value = 0;
	view->progress_last_value = 0;

	status_item_view_progress_set_percent (item, 0, view);
}

static void
status_item_view_progress_next_step (StatusItem* item, StatusItemView* view)
{
	g_assert (IS_STATUS_ITEM(item));
	g_assert (IS_STATUS_ITEM_VIEW(view));

	if (view->steps) {
		gint of_100 = (int)((100.0 * ++view->step)/view->steps);
		status_item_view_progress_set_percent (item, of_100, view);
	}
}

/**
***
**/

static gint
set_progress_activity_mainthread (gpointer data)
{
	StatusItemView *view = STATUS_ITEM_VIEW(data);
	GtkProgress *progress = GTK_PROGRESS(view->progress_major);

	/* make sure we're in activity mode */
	if (!progress->activity_mode) {
		gtk_progress_set_activity_mode (progress, TRUE);
		gtk_progress_bar_set_discrete_blocks(
			GTK_PROGRESS_BAR(progress), 20);
		gtk_progress_bar_set_activity_step (
			GTK_PROGRESS_BAR(progress), 1);
	}

	/* set the value */
	gtk_progress_set_value (progress, view->step%19);

	return 0;
}

static void
status_item_view_progress_set_activity (StatusItem* item,
                                       gint n,
                                       StatusItemView* view)
{
	view->progress_value = n;
	gui_invoke_in_gui_thread (set_progress_activity_mainthread, view);
}

/**
***
**/

typedef struct
{
	StatusItemView *view;
	gchar *text;
}
FooStruct;

static gint
set_status_mainthread (gpointer data)
{
	FooStruct *foo = (FooStruct*) data;
	StatusItemView *view = foo->view;
	gchar *pch = string_shorten (foo->text);

	gtk_progress_set_format_string (
		GTK_PROGRESS(view->progress_major), pch);

	g_free (pch);
	g_free (foo->text);
	g_free (foo);

	return 0;
}

static void
status_item_view_status (StatusItem* item, const gchar* status, StatusItemView* v)
{
	FooStruct *foo = g_new0 (FooStruct, 1);
	foo->view = v;
	foo->text = g_strdup (status);
	gui_invoke_in_gui_thread (set_status_mainthread, foo);
}

static void
status_item_view_error (StatusItem* item, const gchar* error, StatusItemView* v)
{
	status_item_view_status (item, error, v);
	debug(DEBUG_QUEUE,"Error: %s", error);
	pan_error_dialog (error);
}

static void
status_item_view_done (StatusItem* item, gint status, StatusItemView* v)
{
	// FIXME
}

/**
***
**/

static gchar*
string_shorten (const gchar* original)
{
	const int max_width = 35;
	const gchar *middle = "...";
	const size_t middle_len = 3;
	gchar *clone = NULL;
	size_t len = 0;
	size_t ends = 0;

	if (!original)
		return NULL;

	clone = g_strdup(original);
	len = strlen(original);
	if (len<=max_width)
		return clone;

	ends = (max_width)/2 - middle_len;
	memmove (clone, original, ends);
	memmove (clone+ends, middle, middle_len);
	memmove (clone+ends+middle_len, clone+len-ends, ends+1);
	return clone;
}
