/*
 * Pan - A Newsreader for X
 * Copyright (C) 1999  Pan Development Team (pan@superpimp.org)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

#include <sys/stat.h>
#include <errno.h>
#include <ctype.h>
#include "pan.h"


/*---[ wrap_long_lines ]----------------------------------------------
 * take a really long line and put newlines into the appropriate
 * places so the string is never wider than maxlen
 *--------------------------------------------------------------------*/
gchar*
wrap_long_lines (gchar *string, gint maxlen)
{
	gint i=maxlen, j=0;
	gint length = strlen (string) - maxlen;

	while (j <= length) {
		while ((*(string + j + i) != ' ') &&
		       (*(string + j + i) != '\n'))
			i--;
		*(string + j + i) = '\n';
		j += i;
		i = maxlen;
	}
	
	return string;
}


static int
pan_error_dialog_mainthread (gpointer data)
{
	gchar* str = (gchar*) data;
	gnome_error_dialog (str);
	g_free (str);

	return 0;
}
void
pan_error_dialog (const gchar *format, ...)
{
	va_list args;
	gchar *str = NULL;

	g_return_if_fail (format != NULL);

	va_start (args, format);
	str = g_strdup_vprintf (format, args);
	va_end (args);

	gui_invoke_in_gui_thread (pan_error_dialog_mainthread, str);
}



/*---[ not_implemented ]----------------------------------------------
 * ok dialog popup to notify user that we are so lazy we haven't
 * added their feature yet.
 *--------------------------------------------------------------------*/
void
not_implemented (void)
{
	gnome_ok_dialog (_("Not yet implemented."));
}


void
clist_unselect_all (GtkWidget *clist)
{
	GList *list;
	gint row;

	list = GTK_CLIST(clist)->selection;
	while (list) {
		row = GPOINTER_TO_INT (list->data);
		list = list->next;
		gtk_clist_unselect_row (GTK_CLIST(clist), row, 0);
	}
	g_list_free (list);
}


void
pan_lock (void)
{
	pthread_t thr;

	thr = pthread_self();
	if (thr != Pan.main_t)
		gdk_threads_enter();
}


void
pan_unlock (void)
{
	pthread_t thr;

	thr = pthread_self();
	if (thr != Pan.main_t)
		gdk_threads_leave();
}


/*--------------------------------------------------------------------
 *
 *--------------------------------------------------------------------*/
void
directory_check (const gchar *pathname)
{
	struct stat st;
	gchar *dir;
	gint len, i;

	len = strlen (pathname);

	i = 1;
	while (i <= len) {
		if (pathname[i] == '/' || pathname[i] == '\0') {
			dir = g_strndup (pathname, i);
			if (stat (dir, &st) != 0) {
				int ret;
				ret = mkdir (dir, S_IRUSR | S_IWUSR | S_IXUSR);
				if (ret == -1) {
					pan_error_dialog (_("Couldn't create directory: "),
							  pathname, strerror(errno));
					return;
				}
			}
			g_free (dir);
		}
		i++;
	}
}


int lower_bound (
   const void* key,
   const void* base,
   size_t n,
   size_t size,
   int (*compare)(const void *, const void *),
   gboolean* exact_match )
{
   int low = 0;
   int high = n-1;

   while ( low <= high )
   {
      const int mid = (low+high)/2;
      const int comp = (*compare)(key,(const void*)(((char*)(base))+(mid*size)));
      if ( comp > 0 ) low = mid + 1;
      else if ( comp < 0 ) high = mid - 1;
      else {
         if ( exact_match!=NULL )
            *exact_match = TRUE;
         return mid;
      }
   }

   if ( exact_match!=NULL )
      *exact_match = FALSE;
   return low;
}


void
widget_set_font (GtkWidget* w,
		 const char* font_name)
{
	GtkStyle* style = NULL;
	GdkFont* font = NULL;

	g_return_if_fail ( w!=NULL );
	if ( !font_name || !*font_name )
		return;

	style = gtk_style_copy ( gtk_widget_get_style ( Pan.window ) );
	font = gdk_fontset_load ( (char*)font_name );
	if ( font != NULL )
		style->font = font;
	else
		g_warning ( "Couldn't load font ``%s''", font_name );
	gtk_widget_set_style ( w, style );
}


GArray*
read_file (const char* filename)
{
	GArray* array = NULL;
	FILE* fp;

	/* get length of file... */
	size_t len = get_filesize(filename);
	if ( !len )
		return NULL;

	/* open the file... */
	fp = fopen ( filename, "r" );
	if ( !fp ) {
		pan_error_dialog ( _("Unable to open file: %s\n%s"), filename, strerror(errno) );
		return NULL;
	}

	/* read the file... */
	array = g_array_new ( TRUE, FALSE, 1 );
	g_array_set_size ( array, len );
	array->len = fread ( array->data, 1, len, fp );
	fclose ( fp );
	return array;
}


/*--------------------------------------------------------------------
 * 
 *--------------------------------------------------------------------*/
void
commatize_ulong (gulong num,
		 gchar* setme)
{
	char buf[16], *src=buf;
	int len;
	int i;

	sprintf ( buf, "%ld", num );
	len = strlen ( buf );
	i = len % 3;
	for ( ;; ) {
		while ( *src && i-- )
			*setme++ = *src++;
		if ( !*src ) {
			*setme = '\0';
			return;
		}
		else if ( src!=buf )
		   *setme++ = ',';
		i = 3;
	}

	g_assert_not_reached ( );
}

/*---[ get_filesize ]-------------------------------------------------------
 * Use stat() to get and then return the size of the file.  Much faster
 * than opening the file and using ftell().
 *--------------------------------------------------------------------------*/
extern size_t
get_filesize (const gchar* filename)
{
	struct stat buf;
	if ( !stat(filename, &buf) )
		return buf.st_size;

	g_warning ( "Can't stat file ``%s'': %s", filename, strerror(errno) );
	return 0;
}


/**
 * dialog_draw_centered:
 * @dialog: the dialog to draw
 * @window: the window to center the dialog w.r.t.
 *
 * Draw a dialog at the center of a window.
 **/
void
dialog_draw_centered (GtkWidget *dialog, GtkWidget *window)
{
	int x=0, y=0, width=640, height=400;
	int dlg_width, dlg_height;
	int dlg_x, dlg_y;

	if ( !window )
		window = Pan.window;

	gdk_window_get_position (GTK_WIDGET (window)->window, &x, &y);
	gdk_window_get_size (GTK_WIDGET (window)->window, &width, &height);

	gtk_widget_realize (dialog);
	gdk_window_get_size (dialog->window, &dlg_width, &dlg_height);

	dlg_x = x + width/2 - dlg_width/2;
	dlg_y = y + height/2 - dlg_height/2;

	gtk_widget_set_uposition (dialog, dlg_x, dlg_y);
	gtk_widget_show_all (dialog);

	gdk_window_raise (dialog->window);
}
